// SPDX-License-Identifier: GPL-2.0
/*
 * Hidraw Userspace Example
 *
 * Copyright (c) 2010 Alan Ott <alan@signal11.us>
 * Copyright (c) 2010 Signal 11 Software
 *
 * The code may be used by anyone for any purpose,
 * and can serve as a starting point for developing
 * applications using hidraw.
 */
/* C */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
/* Unix */
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h> 
/* Linux */
#include <linux/types.h>
#include <linux/input.h>
#include <linux/hidraw.h>

#include "ioctl.h"
/*
 * Ugly hack to work around failing compilation on systems that don't
 * yet populate new version of hidraw.h to userspace.
 */
#ifndef HIDIOCSFEATURE
#warning Please have your distro update the userspace kernel headers
#define HIDIOCSFEATURE(len)    _IOC(_IOC_WRITE|_IOC_READ, 'H', 0x06, len)
#define HIDIOCGFEATURE(len)    _IOC(_IOC_WRITE|_IOC_READ, 'H', 0x07, len)
#endif

#ifndef USB_RETRY_COUNT
#define USB_RETRY_COUNT 10
#endif

const unsigned retry = USB_RETRY_COUNT;

#include "platform.h"

const char *bus_str(int bus);
static int fd = -1;
static int devtype = 0;
static int UserSpiDataExchangeUSB(uint8_t *pTxBuf, uint8_t *pRxBuf, uint32_t u32PackSize);
static int SysInitUSB(void);

int UserSys_Init(uint32_t u32SpiFreq)
{
        if(devtype == 0){
                return SysInitUSB();
        }
        return -1;
}

static int SysInitUSB(void){
	int i, res, desc_size = 0;
	char buf[256+5];
	//struct hidraw_report_descriptor rpt_desc;
	struct hidraw_devinfo info;
	//char *device = "/dev/hidraw4";
        DIR *dir;
        struct dirent *entry;

        dir = opendir("/dev");
        if (dir) {
                while ((entry = readdir(dir)) != NULL) {
                        if(strncmp( entry->d_name, "hidraw", 6 ) == 0){
                                sprintf( buf, "/dev/%s", entry->d_name);
                                fd = open( buf, O_RDWR|O_NONBLOCK);
	                        if (fd < 0) {
		                        perror("Unable to open device");
		                        continue;
	                        }

	                        //memset(&rpt_desc, 0x0, sizeof(rpt_desc));
	                        memset(&info, 0x0, sizeof(info));
	                        //memset(buf, 0x0, sizeof(buf));
                        	res = ioctl(fd, HIDIOCGRAWINFO, &info);
	                        if (res < 0) {
		                        perror("HIDIOCGRAWINFO");
                                        continue;
	                        } else {
		                        printf("%s Raw Info:\n", buf);
		                        printf("\tbustype: %d (%s)\n",
			                        info.bustype, bus_str(info.bustype));
		                        printf("\tvendor: 0x%04hx\n", info.vendor);
		                        printf("\tproduct: 0x%04hx\n", info.product);
                                        if( (info.bustype == BUS_USB) && 
                                            (info.vendor == 0x16c0) && 
                                            (info.product == 0x2020)){
                                                printf("Find ECM XFU device %s\n", entry->d_name);
                                                break;
                                        }
                                        close(fd);
                                        fd = -1;
	                        }
                        }
                }
                closedir(dir);
        }
        if(fd < 0){
                return -ENXIO;
        }
        printf("HID fd = %d\n", fd);
	return 0;
}

void UserSys_Close(void){
        if(fd >= 0)
                close(fd);
        fd = -1;
}

int UserSpiDataExchange(uint8_t *pTxBuf, uint8_t *pRxBuf, uint32_t u32PackSize){
        if(devtype == 0){
                return UserSpiDataExchangeUSB(pTxBuf, pRxBuf, u32PackSize);
        }
        return -1;
}

static int UserSpiDataExchangeUSB(uint8_t *pTxBuf, uint8_t *pRxBuf, uint32_t u32PackSize){
        int res;
        int i;
        if(fd >= 0){
	        res = write(fd, pTxBuf, u32PackSize);
                for( i = retry; i > 0; i--){
	                if (res < 0) {
		                //printf("Error: %d\n", errno);
		                //perror("write");
                                UserDelay(100);
	                } else {
                                //for(i = 0; i < 24; i+=4)
                                //        printf("%08x ", *(uint32_t *)(&pTxBuf[i]));
		                //printf("write() wrote %d bytes\n", res);
                                break;
	                }
                }
                if(res < 0){
                        return res;
                }
                for( i = retry; i > 0; i--){
	                res = read(fd, pRxBuf, u32PackSize);
	                if (res < 0) {
		                //perror("read");
                                UserDelay(100);         
	                } else {
                                int j;
                                //for(j = 0; j < 4; j+=4)
                                //        printf("%08x ", *(uint32_t *)(&pRxBuf[j]));
		                //printf("read() read %d bytes:\n", res);
                                break;
	                }
                }
        }else{
                return -1;
        }
        return res;
}

int UserDelay(uint32_t MircoSec){
        struct timespec ts;
        ts.tv_sec = MircoSec/1000000;
        ts.tv_nsec = (MircoSec%1000000)*1000;
        return nanosleep(&ts, NULL);
}

const char *
bus_str(int bus)
{
	switch (bus) {
	case BUS_USB:
		return "USB";
		break;
	case BUS_HIL:
		return "HIL";
		break;
	case BUS_BLUETOOTH:
		return "Bluetooth";
		break;
	case BUS_VIRTUAL:
		return "Virtual";
		break;
	default:
		return "Other";
		break;
	}
}

int XF_SetDevType(int nType){
        if(fd >= 0)
                return -EBUSY;
        devtype = nType;
        return 0;
}
int XF_GetDevType(){ return devtype; }
int32_t UserGetGpioData_u32(unsigned long *pulVal){
        if(devtype == 0){
                return 0;
        }
        return -1;
}
int32_t UserSetGpioData_u32(unsigned long ulVal){
        if(devtype == 0){
                return 0;
        }
        return -1;
}
int32_t UserGetGpioDir(unsigned long *pulDir){
        if(devtype == 0){
                return 0;
        }
        return -1;
}
int32_t UserSetGpioDir(unsigned long ulDir){
        if(devtype == 0){
                return 0;
        }
        return -1;
}
int32_t UserSetGpioData(uint8_t u8Idx, unsigned long ulVal){
        unsigned long val;
        int ret;
        if(devtype == 0){
                return 0;
        }
        return -1;
}
int32_t UserGetGpioData(uint8_t u8Idx, unsigned long *pulVal){
        unsigned long val;
        int ret;
        if(devtype == 0){
                return 0;
        }
        return -1; 
}

